from setuptools import Extension
from setuptools.command.build_ext import build_ext as _build_ext
import sys


# bootstrap numpy
# https://stackoverflow.com/questions/19919905/how-to-bootstrap-numpy-installation-in-setup-py
class build_ext(_build_ext):
    def finalize_options(self):
        _build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        #__builtins__.__NUMPY_SETUP__ = False
        import numpy
        self.include_dirs.append(numpy.get_include())


ndl_parallel = Extension("pyndl.ndl_parallel", ["pyndl/ndl_parallel.pyx"])
ndl_openmp = Extension("pyndl.ndl_openmp", ["pyndl/ndl_openmp.pyx"],
                       extra_compile_args=['-fopenmp'], extra_link_args=['-fopenmp'])
corr_parallel = Extension("pyndl.correlation_openmp", ["pyndl/correlation_openmp.pyx"],
                       extra_compile_args=['-fopenmp'], extra_link_args=['-fopenmp'])
# by giving ``cython`` as ``install_requires`` this will be ``cythonized``
# automagically

ext_modules = []
if sys.platform.startswith('linux'):
    ext_modules = [ndl_parallel, ndl_openmp, corr_parallel]
elif sys.platform.startswith('win32'):
    ext_modules = [ndl_parallel] # skip openmp installation on windows for now
elif sys.platform.startswith('darwin'):
    ext_modules = [ndl_parallel]  # skip openmp installation on macos for now


def build(setup_kwargs):
    """
    This function is mandatory in order to build the extensions.
    """
    setup_kwargs.update({
        'ext_modules': ext_modules,
        'cmdclass': {
            'build_ext': build_ext
        }
    })



"""
pyndl.io
--------

*pyndl.io* provides functions to create event generators from different
sources in order to use them with *pyndl.ndl* to train NDL models or to save
existing events from a DataFrame or a list to a file.

"""

import gzip
from collections.abc import Iterator, Iterable
from pathlib import Path
import itertools
import warnings

import pandas as pd


def events_from_file(event_path, compression="gzip", start=0, step=1):
    """
    Yields events for all events in a gzipped event file.

    Parameters
    ----------
    event_path : str
        path to gzipped event file
    compression : str
        indicates whether the events should be read from gunzip
        file or not can be {"gzip" or None}
    start: int
        first event to read
    step: int
        slice every step-th event (useful for parallel computations)

    Yields
    ------
    cues, outcomes : list, list
        a tuple of two lists containing cues and outcomes

    """
    if compression == "gzip":
        event_file = gzip.open(event_path, 'rt')
    elif compression is None:
        event_file = open(event_path, 'rt')
    else:
        raise ValueError("compression needs to be 'gzip' or None")

    try:
        # skip header
        event_file.readline()
        for line in itertools.islice(event_file, start, None, step):
            entries = line.strip('\n').split('\t')
            if len(entries) == 2:
                cues, outcomes = entries
                frequency = 1
            else:
                cues, outcomes, frequency = entries
            cues = cues.split('_')
            outcomes = outcomes.split('_')
            for i in range(int(frequency)):
                yield (cues, outcomes)
    finally:
        event_file.close()


def events_to_file(events, file_path, delimiter="\t", compression="gzip",
                   columns=("cues", "outcomes"), compatible=False):
    """
    Writes events to a file

    Parameters
    ----------
    events : pandas.DataFrame or Iterator or Iterable
        a pandas DataFrame with one event per row and one colum with the cues
        and one column with the outcomes or a list of cues and outcomes as strings
        or a list of a list of cues and a list of outcomes which should be written
        to a file
    file_path: str
        path to where the file should be saved
    delimiter: str
        Seperator which should be used. Default ist a tab
    compression : str
        indicates whether the events should be read from gunzip
        file or not can be {"gzip" or None}
    columns: tuple
        a tuple of column names
    compatible: bool
        if true add a third frequency column (all ones) for compatibility with ndl2
    """
    if isinstance(events, pd.DataFrame):
        events = events_from_dataframe(events)
    elif isinstance(events, (Iterator, Iterable)):
        events = events_from_list(events)
    else:
        raise ValueError("events should either be a pd.DataFrame or an Iterator or an Iterable.")

    if compression == "gzip":
        out_file = gzip.open(file_path, 'wt')
    elif compression is None:
        out_file = open(file_path, 'wt')
    else:
        raise ValueError("compression needs to be 'gzip' or None")

    try:
        legacy_columns = ('Cues', 'Outcomes', 'Frequency')
        if compatible and columns != legacy_columns:
            warnings.warn(f"events_to_file sets the columns to the legacy names '{legacy_columns}' for ndl2 compatibility.\n"
                           "Remove the warning by setting the columns parameter explicitly to this value.")
            columns = legacy_columns
        out_file.write("{}\n".format(delimiter.join(columns)))

        for cues, outcomes in events:
            if isinstance(cues, list) and isinstance(outcomes, list):
                cues = "_".join(cues)
                outcomes = "_".join(outcomes)
            elif not (isinstance(cues, str) and isinstance(outcomes, str)):
                raise ValueError("cues and outcomes should either be a list or a string.")

            if compatible: 
                line = "{}{}{}{}1\n".format(cues, delimiter, outcomes, delimiter)
            else:
                line = "{}{}{}\n".format(cues, delimiter, outcomes)

            out_file.write(line)
    finally:
        out_file.close()


def events_from_dataframe(df, columns=("cues", "outcomes")):
    """
    Yields events for all events in a pandas dataframe.

    Parameters
    ----------
    df : pandas.DataFrame
        a pandas DataFrame with one event per row and one colum with the cues
        and one column with the outcomes.
    columns : tuple
        a tuple of column names

    Yields
    ------
    cues, outcomes : list, list
        a tuple of two lists containing cues and outcomes

    """
    for _, row in df.iterrows():
        cues, outcomes = row[list(columns)]
        cues = cues.split('_')
        outcomes = outcomes.split('_')
        yield (cues, outcomes)


def events_from_list(lst):
    """
    Yields events for all events in a list.

    Parameters
    ----------
    lst : list of list of str or list of str
        a list either containing a list of cues as strings and a list of outcomes
        as strings or a list containing a cue and an outcome string, where cues
        respectively outcomes are seperated by an undescore

    Yields
    ------
    cues, outcomes : list, list
        a tuple of two lists containing cues and outcomes

    """
    for cues, outcomes in lst:
        if isinstance(cues, str):
            cues = cues.split('_')
        if isinstance(outcomes, str):
            outcomes = outcomes.split('_')
        yield (cues, outcomes)


def safe_write_path(path, template='{path.stem}-{counter}{path.suffix}'):
    """
    Create a file path to avoid overwriting existing files.
    Returns the original path if it does not exist or
    an incremented version according to the template.
    
    This function with the default template creates filenames like
    pathname/example.png, pathname/example-1.png, pathname/example-2.png, ...

    Parameters
    ----------
    path: file path
    template: format string syntax of incremented file name.
              available variables are counter (int) and path (pathlib.Path).

    Returns
    -------
    path: the input path or (if file exists) the path with incremented filename.
    """
    if template.format(path=path, counter=1) == template.format(path=path, counter=2):
        raise ValueError(f"Expects template to change by '{{counter}}', got {template}")

    new_path = path = Path(path)
    base_dir = path.parent.resolve()
    counter = 0
    while new_path.exists():
        counter = counter + 1
        new_path = Path(template.format(path=path, counter=counter))

        # is new_path already relative to path's directory?
        if not str(new_path.resolve()).startswith(str(base_dir)):
            new_path = path.parent / new_path

    return new_path



"""
Pyndl - Naive Discriminative Learning in Python
===============================================

*pyndl* implements Naïve Discriminative Learning (NDL) in Python. NDL is an
incremental learning algorithm grounded in the principles of discrimination
learning and motivated by animal and human learning research. Lately, NDL
has become a popular tool in language research to examine large corpora and
vocabularies, with 750,000 spoken word tokens and a vocabulary size of 52,402
word types. In contrast to previous implementations, *pyndl* allows for a
broader range of analysis, including non-English languages, adds further
learning rules and provides better maintainability while having the same
fast processing speed. As of today, it supports multiple research groups
in their work and led to several scientific publications.

"""

import platform
import os
import sys
import multiprocessing as mp
try:
    from importlib.metadata import requires
except ModuleNotFoundError:  # python 3.7 and before
    requires = None
try:
    from packaging.requirements import Requirement
except ModuleNotFoundError:  # this should only happend during setup phase
    Requirement = None

try:
    from importlib import metadata
    __version__ = metadata.version(__package__)
except metadata.PackageNotFoundError:
    import toml
    __version__ = toml.load("pyproject.toml")["tool"]["poetry"]["version"] + "dev"


def sysinfo():
    """
    Prints system the dependency information
    """
    if requires:
        dependencies = [Requirement(req).name for req in requires('pyndl')
                        if not Requirement(req).marker]

    header = ("Pyndl Information\n"
              "=================\n\n")

    general = ("General Information\n"
               "-------------------\n"
               "Python version: {}\n"
               "Pyndl version: {}\n\n").format(sys.version.split()[0], __version__)

    uname = platform.uname()
    osinfo = ("Operating System\n"
              "----------------\n"
              "OS: {s.system} {s.machine}\n"
              "Kernel: {s.release}\n"
              "CPU: {cpu_count}\n").format(s=uname, cpu_count=mp.cpu_count())

    if uname.system == "Linux":
        _, *lines = os.popen("free -m").readlines()
        for identifier in ("Mem:", "Swap:"):
            memory = [line for line in lines if identifier in line]
            if len(memory) > 0:
                _, total, used, *_ = memory[0].split()
            else:
                total, used = '?', '?'
            osinfo += "{} {}MiB/{}MiB\n".format(identifier, used, total)

    osinfo += "\n"

    deps = ("Dependencies\n"
            "------------\n")

    if requires:
        deps += "\n".join("{pkg.__name__}: {pkg.__version__}".format(pkg=__import__(dep))
                          for dep in dependencies)
    else:
        deps = 'You need Python 3.8 or higher to show dependencies.'

    print(header + general + osinfo + deps)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyndl']

package_data = \
{'': ['*']}

install_requires = \
['Cython>=0.29.32,<0.30.0',
 'netCDF4>=1.6.0,<2.0.0',
 'numpy>=1.23.1,<2.0.0',
 'pandas>=1.4.3,<2.0.0',
 'scipy>=1.9.0,<2.0.0',
 'xarray>=2022.6.0,<2023.0.0']

setup_kwargs = {
    'name': 'pyndl',
    'version': '1.1.1',
    'description': 'Naive discriminative learning implements learning and classification models based on the Rescorla-Wagner equations.',
    'long_description': "===============================================\nPyndl - Naive Discriminative Learning in Python\n===============================================\n\n.. image:: https://github.com/quantling/pyndl/actions/workflows/python-test.yml/badge.svg?branch=main\n    :target: https://github.com/quantling/pyndl/actions/workflows/python-test.yml\n\n.. image:: https://codecov.io/gh/quantling/pyndl/branch/main/graph/badge.svg?token=2GWUXRA9PD\n    :target: https://codecov.io/gh/quantling/pyndl\n\n.. image:: https://img.shields.io/lgtm/grade/python/g/quantling/pyndl.svg?logo=lgtm&logoWidth=18\n    :target: https://lgtm.com/projects/g/quantling/pyndl/context:python\n\n.. image:: https://img.shields.io/pypi/pyversions/pyndl.svg\n    :target: https://pypi.python.org/pypi/pyndl/\n\n.. image:: https://img.shields.io/github/license/quantling/pyndl.svg\n    :target: https://github.com/quantling/pyndl/blob/main/LICENSE\n\n.. image:: https://zenodo.org/badge/DOI/10.5281/zenodo.597964.svg\n   :target: https://doi.org/10.5281/zenodo.597964\n\n*pyndl* is an implementation of Naive Discriminative Learning in Python. It was\ncreated to analyse huge amounts of text file corpora. Especially, it allows to\nefficiently apply the Rescorla-Wagner learning rule to these corpora.\n\n\nInstallation\n============\n\nThe easiest way to install *pyndl* is using\n`pip <https://pip.pypa.io/en/stable/>`_:\n\n.. code:: bash\n\n    pip install --user pyndl\n\nFor more information have a look at the `Installation Guide\n<http://pyndl.readthedocs.io/en/latest/installation.html>`_.\n\n\nDocumentation\n=============\n\n*pyndl* uses ``sphinx`` to create a documentation manual. The documentation is\nhosted on `Read the Docs <http://pyndl.readthedocs.io/en/latest/>`_.\n\n\nGetting involved\n================\n\nThe *pyndl* project welcomes help in the following ways:\n\n* Making Pull Requests for\n  `code <https://github.com/quantling/pyndl/tree/main/pyndl>`_,\n  `tests <https://github.com/quantling/pyndl/tree/main/tests>`_\n  or `documentation <https://github.com/quantling/pyndl/tree/main/doc>`_.\n* Commenting on `open issues <https://github.com/quantling/pyndl/issues>`_\n  and `pull requests <https://github.com/quantling/pyndl/pulls>`_.\n* Helping to answer `questions in the issue section\n  <https://github.com/quantling/pyndl/labels/question>`_.\n* Creating feature requests or adding bug reports in the `issue section\n  <https://github.com/quantling/pyndl/issues/new>`_.\n\nFor more information on how to contribute to *pyndl* have a look at the\n`development section <http://pyndl.readthedocs.io/en/latest/development.html>`_.\n\n\nAuthors and Contributers\n========================\n\n*pyndl* was mainly developed by\n`Konstantin Sering <https://github.com/derNarr>`_,\n`Marc Weitz <https://github.com/trybnetic>`_,\n`David-Elias Künstle <https://github.com/dekuenstle/>`_,\n`Elnaz Shafaei Bajestan <https://github.com/elnazsh>`_\nand `Lennart Schneider <https://github.com/sumny>`_. For the full list of\ncontributers have a look at `Github's Contributor summary\n<https://github.com/quantling/pyndl/contributors>`_.\n\nCurrently, it is maintained by `Konstantin Sering <https://github.com/derNarr>`_\nand `Marc Weitz <https://github.com/trybnetic>`_.\n\n\nFunding\n-------\n*pyndl* was partially funded by the Humboldt grant, the ERC advanced grant (no.\n742545) and by the University of Tübingen.\n\n\nAcknowledgements\n----------------\nThis package is build as a python replacement for the R ndl2 package. Some\nideas on how to build the API and how to efficiently run the Rescorla Wagner\niterative learning on large text corpora are inspired by the way the ndl2\npackage solves this problems. The ndl2 package is available on Github `here\n<https://github.com/quantling/ndl2>`_.\n\n",
    'author': 'Konstantin Sering',
    'author_email': 'konstantin.sering@uni-tuebingen.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://pyndl.readthedocs.io/en/latest/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.12',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



"""
pyndl.wh
--------

*pyndl.wh* provides functions in order to train Widrow-Hoff (WH) models. In contrast
to the Rescorla-Wagner (RW) models, the WH models can not only have binary
cues and outcomes, but can encode gradual intensities in the cues and outcomes.
This is done by associating a vector of continues values (real numbers) to each
cue and outcome. The size of the vector has to be the same for all cues and for
all outcomes, but can differ between cues and outcomes.

It is possible to calculate weights for continuous cues or continues outcomes,
while keeping the outcomes respectively cues binary. Finally, it is possible to
have both sides, cues and outcomes, to be continues and calculate the
Widrow-Hoff learning rule between them.

"""
from collections import defaultdict, OrderedDict
import copy
import getpass
import os
# from queue import Queue
import socket
import sys
import tempfile
# import threading
import time

import cython
import pandas as pd
import numpy as np
import xarray as xr

from . import __version__ as pyndl_version
from . import count
from . import preprocess
from . import io
from . import ndl

# conditional import as openmp is only compiled for linux
if sys.platform.startswith('linux'):
    from . import ndl_openmp
elif sys.platform.startswith('win32'):
    pass
elif sys.platform.startswith('darwin'):
    pass


WeightDict = ndl.WeightDict


def wh(events, eta, *, cue_vectors=None, outcome_vectors=None,
        method='openmp', weights=None,
        n_jobs=8, n_outcomes_per_job=10, remove_duplicates=None,
        verbose=False, temporary_directory=None,
        events_per_temporary_file=10000000):
    """
    Calculate the weights for all events using the Widrow-Hoff learning rule in
    three different flavors.

    In the first flavor, cues and outcomes both are vectors and the names in
    the eventfiles refer to these vectors. The vectors for all cues and
    outcomes are given as an xarray.DataArray with the arguments `cue_vectors`
    and `outcome_vectors'.

    In the second and third flavor, only the cues or only the outcomes are
    treated as vectors and the ones not being treated as vectors are still
    considered being present or not being present in a binary way.

    This is a parallel python implementation using cython, numpy,
    multithreading and the binary format defined in preprocess.py.

    Parameters
    ----------
    events : str
        path to the event file
    eta : float
        learning rate
    cue_vectors : xarray.DataArray
        matrix that contains the cue vectors for each cue
    outcome_vectors : xarray.DataArray
        matrix that contains the target vectors for each outcome
    method : {'openmp', 'threading', 'numpy'}
        'numpy' works only for real to real Widrow-Hoff.
    weights : None or xarray.DataArray
        the xarray.DataArray needs to have the named dimensions 'cues' or
        'cue_vector_dimensions' and 'outcomes' or 'outcome_vector_dimensions'
    n_jobs : int
        an integer giving the number of threads in which the job should be
        executed
    n_outcomes_per_job : int
        an integer giving the number of outcomes that are processed in one job
    remove_duplicates : {None, True, False}
        if None raise a ValueError when the same cue is present multiple times
        in the same event; True make cues and outcomes unique per event; False
        keep multiple instances of the same cue or outcome (this is usually not
        preferred!)
    verbose : bool
        print some output if True
    temporary_directory : str
        path to directory to use for storing temporary files created;
        if none is provided, the operating system's default will
        be used like '/tmp' on unix
    events_per_temporary_file: int
        Number of events in each temporary binary file. Has to be larger than 1

    Returns
    -------
    weights : xarray.DataArray
        the dimensions of the weights reflect the type of Widrow-Hoff that was
        run (real to real, binary to real, real to binary or binary to binary).
        The dimension names reflect this in the weights. They are a combination
        of 'outcomes' x 'outcome_vector_dimensions' and 'cues' x
        'cue_vector_dimensions' with dimensions 'outcome_vector dimensions' and
        'cue_vector_dimensions'. You can lookup the weights between a vector
        dimension and a cue with ``weights.loc[{'outcome_vector_dimensions':
        outcome_vector_dimension, 'cue_vector_dimensions':
        cue_vector_dimension}]`` or
        ``weights.loc[vector_dimension].loc[cue_vector_dimension]``.

    """
    if cue_vectors is None and outcome_vectors is None:
        lambda_ = 1.0
        alpha = 1.0
        betas = (eta, eta)
        return ndl.ndl(events, alpha, betas, lambda_,
                       method=method, weights=weights, n_jobs=n_jobs,
                       n_outcomes_per_job=n_outcomes_per_job,
                       remove_duplicates=remove_duplicates, verbose=verbose,
                       temporary_directory=temporary_directory,
                       events_per_temporary_file=events_per_temporary_file)
    elif cue_vectors is not None and outcome_vectors is not None:
        return _wh_real_to_real(events, eta, cue_vectors, outcome_vectors,
                                method=method, weights=weights, n_jobs=n_jobs,
                                n_outcomes_per_job=n_outcomes_per_job,
                                remove_duplicates=remove_duplicates, verbose=verbose,
                                temporary_directory=temporary_directory,
                                events_per_temporary_file=events_per_temporary_file)
    elif cue_vectors is not None and outcome_vectors is None:
        lambda_ = 1.0
        betas = (eta, eta)
        return _wh_real_to_binary(events, betas, lambda_, cue_vectors,
                                  method=method, weights=weights, n_jobs=n_jobs,
                                  n_outcomes_per_job=n_outcomes_per_job,
                                  remove_duplicates=remove_duplicates, verbose=verbose,
                                  temporary_directory=temporary_directory,
                                  events_per_temporary_file=events_per_temporary_file)
    elif cue_vectors is None and outcome_vectors is not None:
        return _wh_binary_to_real(events, eta, outcome_vectors,
                                  method=method, weights=weights, n_jobs=n_jobs,
                                  n_outcomes_per_job=n_outcomes_per_job,
                                  remove_duplicates=remove_duplicates, verbose=verbose,
                                  temporary_directory=temporary_directory,
                                  events_per_temporary_file=events_per_temporary_file)
    # The if statements above are covering all cases.


def dict_wh(events, eta, cue_vectors, outcome_vectors, *,
            weights=None, inplace=False, remove_duplicates=None,
            make_data_array=False, verbose=False):
    """
    Calculate the weights for all_outcomes over all events in events.

    This is a pure python implementation using dicts.

    Notes
    -----
    The metadata will only be stored when `make_data_array` is True and then
    `dict_ndl` cannot be used to continue learning. At the moment there is no
    proper way to automatically store the meta data into the default dict.

    Furthermore, this implementation only supports the 'real to real' case
    where cue vectors are learned on outcome vectors. For the 'binary to real'
    or 'real to binary' cases the `wh.wh` function needs to be used which uses
    a fast cython implementation.

    The main purpose of this function is to have a reference implementation
    which is used to validate the faster cython version against. Additionally, this
    function can be a good starting point to develop different flavors of the
    Widrow-Hoff learning rule.

    Parameters
    ----------
    events : generator or str
        generates cues, outcomes pairs or the path to the event file
    eta : float
        learning rate
    cue_vectors : xarray.DataArray
        matrix that contains the cue vectors for each cue
    outcome_vectors : xarray.DataArray
        matrix that contains the target vectors for each outcome
    weights : dict of dicts or xarray.DataArray or None
        initial weights
    inplace: {True, False}
        if True calculates the weightmatrix inplace
        if False creates a new weightmatrix to learn on
    remove_duplicates : {None, True, False}
        if None though a ValueError when the same cue is present multiple times
        in the same event; True make cues and outcomes unique per event; False
        keep multiple instances of the same cue or outcome (this is usually not
        preferred!)
    make_data_array : {False, True}
        if True makes a xarray.DataArray out of the dict of dicts.
    verbose : bool
        print some output if True.

    Returns
    -------
    weights : dict of dicts of floats
        the first dict has outcomes as keys and dicts as values
        the second dict has cues as keys and weights as values
        weights[outcome][cue] gives the weight between outcome and cue.

    or

    weights : xarray.DataArray
        with dimensions 'outcome_vector_dimensions' and
        'cue_vector_dimensions'. You can lookup the weights
        between a cue dimension and an outcome dimension with
        ``weights.loc[{'outcome_vector_dimensions': outcome_vector_dimension,
        'cue_vector_dimensions': cue_vector_dimension}]`` or
        ``weights.loc[outcome_vector_dimension].loc[cue_vector_dimension]``.

    """

    if not isinstance(make_data_array, bool):
        raise ValueError("make_data_array must be True or False")

    if not (remove_duplicates is None or isinstance(remove_duplicates, bool)):
        raise ValueError("remove_duplicates must be None, True or False")

    wall_time_start = time.perf_counter()
    cpu_time_start = time.process_time()
    if isinstance(events, str):
        event_path = events
    else:
        event_path = ""
    attrs_to_update = None

    # weights can be seen as an infinite outcome by cue matrix
    # weights[outcome][cue]
    if weights is None:
        weights = WeightDict()
    elif isinstance(weights, WeightDict):
        attrs_to_update = weights.attrs
    elif isinstance(weights, xr.DataArray):
        raise NotImplementedError('initilizing with a xr.DataArray is not supported yet.')
        # weights_ini = weights
        # attrs_to_update = weights_ini.attrs
        # coords = weights_ini.coords
        # weights = WeightDict()
        # for outcome_index, outcome in enumerate(coords['outcomes'].values):
        #     for cue_index, cue in enumerate(coords['cues'].values):
        #         weights[outcome][cue] = weights_ini.item((outcome_index, cue_index))
    elif not isinstance(weights, defaultdict):
        raise ValueError('weights needs to be either defaultdict or None')

    if not inplace:
        weights = copy.deepcopy(weights)

    if isinstance(events, str):
        events = io.events_from_file(events)
    number_events = 0

    for cues, outcomes in events:
        number_events += 1
        if verbose and number_events % 1000:
            print('.', end='')
            sys.stdout.flush()
        if remove_duplicates is None:
            if (len(cues) != len(set(cues)) or
                    len(outcomes) != len(set(outcomes))):
                raise ValueError('cues or outcomes needs to be unique: cues '
                                 '"%s"; outcomes "%s"; use '
                                 'remove_duplicates=True' %
                                 (' '.join(cues), ' '.join(outcomes)))
        elif remove_duplicates:
            cues = set(cues)
            outcomes = set(outcomes)
        else:
            pass

        assert len(outcomes) == 1, 'for real_wh only one outcome is allowed per event'
        assert len(cues) == 1, 'for real_wh only one cue is allowed per event'

        cue_vec = cue_vectors.loc[cues[0]]
        outcome_vec = outcome_vectors.loc[outcomes[0]]

        for outcome_index in outcome_vec.coords['outcome_vector_dimensions']:
            outcome_index = str(outcome_index.values)
            prediction_strength = 0
            for cue_index in cue_vec.coords['cue_vector_dimensions']:
                cue_index = str(cue_index.values)
                prediction_strength += float(cue_vec.loc[cue_index] * weights[outcome_index][cue_index])
            error = float(outcome_vec.loc[outcome_index] - prediction_strength)

            for cue_index in cue_vec.coords['cue_vector_dimensions']:
                cue_index = str(cue_index.values)
                weights[outcome_index][cue_index] += float(eta * error * cue_vec.loc[cue_index])

    cpu_time_stop = time.process_time()
    wall_time_stop = time.perf_counter()
    cpu_time = cpu_time_stop - cpu_time_start
    wall_time = wall_time_stop - wall_time_start
    attrs = _attributes(event_path, number_events, eta, cpu_time, wall_time,
                        __name__ + "." + dict_wh.__name__, attrs=attrs_to_update)

    if make_data_array:
        outcome_dims = list(weights.keys())
        cue_dims = set()
        for outcome_dim in outcome_dims:
            cue_dims.update(set(weights[outcome_dim].keys()))

        cue_dims = list(cue_dims)

        weights_dict = weights
        shape = (len(outcome_dims), len(cue_dims))
        weights = xr.DataArray(np.zeros(shape), attrs=attrs,
                               coords={'outcome_vector_dimensions': outcome_dims,
                                       'cue_vector_dimensions': cue_dims},
                               dims=('outcome_vector_dimensions', 'cue_vector_dimensions'))

        for outcome_dim in outcome_dims:
            for cue_dim in cue_dims:
                weights.loc[{"outcome_vector_dimensions": outcome_dim,
                             "cue_vector_dimensions": cue_dim}] = weights_dict[outcome_dim][cue_dim]
    else:
        weights.attrs = attrs

    return weights


def _wh_binary_to_real(events, eta, outcome_vectors, *,
                       method='openmp', weights=None,
                       n_jobs=8, n_outcomes_per_job=10, remove_duplicates=None,
                       verbose=False, temporary_directory=None,
                       events_per_temporary_file=10000000):
    """
    Calculate the weights for all events using the Widrow-Hoff learning rule
    from binary cues onto continuous outcome vectors.

    This is a parallel python implementation using cython, numpy,
    multithreading and the binary format defined in preprocess.py.

    Parameters
    ----------
    events : str
        path to the event file
    eta : float
        learning rate
    outcome_vectors : xarray.DataArray
        matrix that contains the target vectors for each outcome

    method : {'openmp', 'threading'}
    weights : None or xarray.DataArray
        the xarray.DataArray needs to have the dimensions 'cues' and
        'outcome_vector_dimensions'
    n_jobs : int
        an integer giving the number of threads in which the job should be
        executed
    n_outcomes_per_job : int
        an integer giving the number of outcomes that are processed in one job
    remove_duplicates : {None, True, False}
        if None though a ValueError when the same cue is present multiple times
        in the same event; True make cues and outcomes unique per event; False
        keep multiple instances of the same cue or outcome (this is usually not
        preferred!)
    verbose : bool
        print some output if True.
    temporary_directory : str
        path to directory to use for storing temporary files created;
        if none is provided, the operating system's default will
        be used like '/tmp' on unix
    events_per_temporary_file: int
        Number of events in each temporary binary file. Has to be larger than 1

    Returns
    -------
    weights : xarray.DataArray
        with dimensions 'outcome_vector dimensions' and 'cues'. You can lookup
        the weights between a vector dimension and a cue with
        ``weights.loc[{'outcome_vector_dimensions': outcome_vector_dimension,
        'cues': cue}]`` or ``weights.loc[outcome_vector_dimension].loc[cue]``.

    """

    if not (remove_duplicates is None or isinstance(remove_duplicates, bool)):
        raise ValueError("remove_duplicates must be None, True or False")
    if not isinstance(events, str):
        raise ValueError("'events' need to be the path to a gzipped event file not {}".format(type(events)))

    weights_ini = weights
    wall_time_start = time.perf_counter()
    cpu_time_start = time.process_time()

    if type(outcome_vectors) == dict:
        # TODO: convert dict to xarray here
        raise NotImplementedError('dicts are not supported yet.')

    if not outcome_vectors.data.data.c_contiguous:
        raise ValueError('outcome_vectors have to be c_contiguous')
    if not outcome_vectors.dtype == np.float64:
        raise ValueError('outcome_vectors have to be of dtype np.float64')

    # preprocessing
    n_events, cues, outcomes_from_events = count.cues_outcomes(events,
                                                               n_jobs=n_jobs,
                                                               verbose=verbose)
    cues = list(cues.keys())
    outcomes_from_events = list(outcomes_from_events.keys())
    outcomes = outcome_vectors.coords['outcomes'].values.tolist()
    cue_map = OrderedDict(((cue, ii) for ii, cue in enumerate(cues)))
    outcome_map = OrderedDict(((outcome, ii) for ii, outcome in enumerate(outcomes)))

    # check for unseen outcomes in events
    if set(outcomes_from_events) - set(outcomes):
        raise ValueError("all outcomes in events need to be specified as rows in outcome_vectors")

    if weights is not None and weights.shape[0] != outcome_vectors.shape[1]:
        raise ValueError("outcome dimensions in weights need to match dimensions in outcome_vectors")

    shape = (outcome_vectors.shape[1], len(cue_map))

    # initialize weights
    if weights is None:
        weights = np.ascontiguousarray(np.zeros(shape, dtype=np.float64, order='C'))
    elif isinstance(weights, xr.DataArray):
        outcome_vector_dimensions = outcome_vectors.coords["outcome_vector_dimensions"].values.tolist()
        if not outcome_vector_dimensions == weights.coords["outcome_vector_dimensions"].values.tolist():
            raise ValueError('Outcome vector dimensions in weights and outcome_vectors do not match!')

        old_cues = weights.coords["cues"].values.tolist()
        new_cues = list(set(cues) - set(old_cues))
        cues = old_cues + new_cues
        cue_map = OrderedDict(((cue, ii) for ii, cue in enumerate(cues)))

        weights_tmp = np.concatenate((weights.values,
                                      np.zeros((len(outcome_vector_dimensions), len(new_cues)),
                                               dtype=np.float64, order='C')),
                                     axis=1)
        weights = np.ascontiguousarray(weights_tmp)

        del weights_tmp, old_cues, new_cues, outcome_vector_dimensions
    else:
        raise ValueError('weights need to be None or xarray.DataArray with method=%s' % method)

    with tempfile.TemporaryDirectory(prefix="pyndl", dir=temporary_directory) as binary_path:
        number_events = preprocess.create_binary_event_files(events, binary_path, cue_map,
                                                             outcome_map, overwrite=True,
                                                             n_jobs=n_jobs,
                                                             events_per_file=events_per_temporary_file,
                                                             remove_duplicates=remove_duplicates,
                                                             verbose=verbose)
        assert n_events == number_events, (str(n_events) + ' ' + str(number_events))
        binary_files = [os.path.join(binary_path, binary_file)
                        for binary_file in os.listdir(binary_path)
                        if os.path.isfile(os.path.join(binary_path, binary_file))]
        # sort binary files as they were created
        binary_files.sort(key=lambda filename: int(os.path.basename(filename)[9:-4]))
        if verbose:
            print('start learning...')
        # learning
        if not weights.data.c_contiguous:
            raise ValueError('weights has to be c_contiguous')
        if method == 'openmp':
            if not sys.platform.startswith('linux'):
                raise NotImplementedError("OpenMP is linux only at the moment."
                                          "Use method='threading' instead.")
            ndl_openmp.learn_inplace_binary_to_real(binary_files,
                                                    eta,
                                                    outcome_vectors.data,
                                                    weights,
                                                    n_outcomes_per_job,
                                                    n_jobs)
        elif method == 'threading':
            raise ValueError('TODO: for now: method needs "openmp"')
        #    part_lists = ndl.slice_list(all_outcome_indices, n_outcomes_per_job)

        #    working_queue = Queue(len(part_lists))
        #    threads = []
        #    queue_lock = threading.Lock()

        #    def worker():
        #        while True:
        #            with queue_lock:
        #                if working_queue.empty():
        #                    break
        #                data = working_queue.get()
        #            ndl_parallel.learn_inplace_2(binary_files, outcome_vectors,
        #                                         eta, weights, data)

        #    with queue_lock:
        #        for partlist in part_lists:
        #            working_queue.put(np.array(partlist, dtype=np.uint32))

        #    for _ in range(n_jobs):
        #        thread = threading.Thread(target=worker)
        #        thread.start()
        #        threads.append(thread)

        #    for thread in threads:
        #        thread.join()
        else:
            raise ValueError('method needs to be either "openmp"')

    cpu_time_stop = time.process_time()
    wall_time_stop = time.perf_counter()
    cpu_time = cpu_time_stop - cpu_time_start
    wall_time = wall_time_stop - wall_time_start

    if weights_ini is not None:
        attrs_to_be_updated = weights_ini.attrs
    else:
        attrs_to_be_updated = None

    attrs = _attributes(events, number_events, eta, cpu_time, wall_time,
                        __name__ + "." + ndl.__name__, method=method, attrs=attrs_to_be_updated)

    # post-processing
    weights = xr.DataArray(weights, coords=[('outcome_vector_dimensions',
                                             outcome_vectors.coords['outcome_vector_dimensions'].data),
                                            ('cues', cues)], attrs=attrs)
    return weights


def _wh_real_to_binary(events, betas, lambda_, cue_vectors, *,
                       method='openmp', weights=None,
                       n_jobs=8, n_outcomes_per_job=10, remove_duplicates=None,
                       verbose=False, temporary_directory=None,
                       events_per_temporary_file=10000000):
    """
    Calculate the weights for all events using the Widrow-Hoff learning rule
    from continuous cue vectors onto binary outcomes.

    This is a parallel python implementation using cython and the binary format
    defined in preprocess.py.

    Parameters
    ----------
    events : str
        path to the event file
    betas : (float, float)
        one value for successful prediction (reward) one for punishment
    lambda\\_ : float
    cue_vectors : xarray.DataArray
        matrix that contains the cue vectors for each cue
    method : {'openmp', 'threading'}
    weights : None or xarray.DataArray
        the xarray.DataArray needs to have the dimensions
        'cue_vector_dimensions' and 'outcomes'
    n_jobs : int
        an integer giving the number of threads in which the job should be
        executed
    n_outcomes_per_job : int
        an integer giving the number of outcomes that are processed in one job
    remove_duplicates : {None, True, False}
        if None though a ValueError when the same cue is present multiple times
        in the same event; True make cues and outcomes unique per event; False
        keep multiple instances of the same cue or outcome (this is usually not
        preferred!)
    verbose : bool
        print some output if True.
    temporary_directory : str
        path to directory to use for storing temporary files created;
        if none is provided, the operating system's default will
        be used like '/tmp' on unix
    events_per_temporary_file: int
        Number of events in each temporary binary file. Has to be larger than 1

    Returns
    -------
    weights : xarray.DataArray
        with dimensions outcomes and cue_vector_dimensions. You can lookup the
        weights between an outcome and a cue_vector_dimension with
        ``weights.loc[{'outcome': outcome, 'cue_vector_dimensions':
        cue_vector_dimension}]`` or
        ``weights.loc[outcome].loc[cue_vector_dimension]``.

    """
    if not (remove_duplicates is None or isinstance(remove_duplicates, bool)):
        raise ValueError("remove_duplicates must be None, True or False")

    if not isinstance(events, str):
        raise ValueError("'events' need to be the path to a gzipped event file not {}".format(type(events)))

    if type(cue_vectors) == dict:
        # TODO: convert dict to xarray here
        raise NotImplementedError('dicts are not supported yet.')

    if not cue_vectors.data.data.c_contiguous:
        raise ValueError('cue_vectors have to be c_contiguous')
    if not cue_vectors.dtype == np.float64:
        raise ValueError('cue_vectors have to be of dtype np.float64')

    weights_ini = weights
    wall_time_start = time.perf_counter()
    cpu_time_start = time.process_time()

    # preprocessing
    n_events, cues_from_events, outcomes_from_events = count.cues_outcomes(events,
                                                                           n_jobs=n_jobs,
                                                                           verbose=verbose)

    cues_from_events = list(cues_from_events.keys())
    cues = list(cue_vectors.coords['cues'].data)
    outcomes = list(outcomes_from_events.keys())

    cue_map = OrderedDict(((cue, ii) for ii, cue in enumerate(cues)))
    outcome_map = OrderedDict(((outcome, ii) for ii, outcome in enumerate(outcomes)))

    if set(cues_from_events) - set(cues):
        raise ValueError("all cues in events need to be specified as rows in cue_vectors")

    del outcomes_from_events, cues_from_events

    shape = (len(outcomes), cue_vectors.shape[1])

    if not cue_vectors.dims[1] == 'cue_vector_dimensions':
        raise ValueError("The second dimension of the 'cue_vectors' has to be named 'cue_vector_dimensions'.")

    cue_vector_dimensions = cue_vectors['cue_vector_dimensions']

    # initialize weights
    if weights is None:
        weights = np.ascontiguousarray(np.zeros(shape, dtype=np.float64, order='C'))
    elif isinstance(weights, xr.DataArray):
        if not all(cue_vector_dimensions == weights['cue_vector_dimensions']):
            raise ValueError("Cue vector dimensions names do not match in weights and cue_vectors")

        old_outcomes = weights.coords["outcomes"].values.tolist()
        new_outcomes = list(set(outcomes) - set(old_outcomes))
        outcomes = old_outcomes + new_outcomes

        outcome_map = OrderedDict(((outcome, ii) for ii, outcome in enumerate(outcomes)))
        weights_tmp = np.concatenate((weights.values,
                                      np.zeros((len(new_outcomes), len(cue_vector_dimensions)),
                                               dtype=np.float64, order='C')),
                                     axis=0)
        weights = np.ascontiguousarray(weights_tmp)
        del weights_tmp, old_outcomes, new_outcomes
    else:
        raise ValueError('weights need to be None or xarray.DataArray with method=%s' % method)

    weights = xr.DataArray(weights,
                           coords={'outcomes': outcomes, 'cue_vector_dimensions': cue_vector_dimensions},
                           dims=('outcomes', 'cue_vector_dimensions'))
    del shape, cue_vector_dimensions

    with tempfile.TemporaryDirectory(prefix="pyndl", dir=temporary_directory) as binary_path:
        number_events = preprocess.create_binary_event_files(events, binary_path, cue_map,
                                                             outcome_map, overwrite=True,
                                                             n_jobs=n_jobs,
                                                             events_per_file=events_per_temporary_file,
                                                             remove_duplicates=remove_duplicates,
                                                             verbose=verbose)
        assert n_events == number_events, (str(n_events) + ' ' + str(number_events))
        binary_files = [os.path.join(binary_path, binary_file)
                        for binary_file in os.listdir(binary_path)
                        if os.path.isfile(os.path.join(binary_path, binary_file))]
        # sort binary files as they were created
        binary_files.sort(key=lambda filename: int(os.path.basename(filename)[9:-4]))
        if verbose:
            print('start learning...')
        # learning
        if not weights.data.data.c_contiguous:
            raise ValueError('weights has to be c_contiguous')
        if method == 'openmp':
            if not sys.platform.startswith('linux'):
                raise NotImplementedError("OpenMP is linux only at the moment."
                                          "Use method='threading' instead.")
            beta1, beta2 = betas
            ndl_openmp.learn_inplace_real_to_binary(binary_files,
                                                    beta1,
                                                    beta2,
                                                    lambda_,
                                                    cue_vectors.data,
                                                    weights.data,
                                                    n_outcomes_per_job,
                                                    n_jobs)
        else:
            # TODO: implement threading
            raise ValueError('TODO: for now: method needs to be "openmp"')

        weights = weights.reset_coords(drop=True)

    cpu_time_stop = time.process_time()
    wall_time_stop = time.perf_counter()
    cpu_time = cpu_time_stop - cpu_time_start
    wall_time = wall_time_stop - wall_time_start

    if weights_ini is not None:
        attrs_to_be_updated = weights_ini.attrs
    else:
        attrs_to_be_updated = None

    attrs = ndl._attributes(events, number_events, 'cue_vectors', betas, lambda_, cpu_time, wall_time,
                            __name__ + "." + ndl.__name__, method=method, attrs=attrs_to_be_updated)

    # post-processing
    weights.attrs = attrs
    return weights


def _wh_real_to_real(events, eta, cue_vectors, outcome_vectors, *,
                     method='openmp', weights=None,
                     n_jobs=8, n_outcomes_per_job=10, remove_duplicates=None,
                     verbose=False, temporary_directory=None,
                     events_per_temporary_file=10000000):
    """
    Calculate the weights for all events using the Widrow-Hoff learning rule
    from cue vectors onto outcome vectors.

    This is a parallel python implementation using cython, numpy,
    multithreading and the binary format defined in preprocess.py.

    Parameters
    ----------
    events : str
        path to the event file
    eta : float
        learning rate
    cue_vectors : xarray.DataArray
        matrix that contains the cue vectors for each cue
    outcome_vectors : xarray.DataArray
        matrix that contains the target vectors for each outcome
    method : {'openmp', 'threading', 'numpy'}
    weights : None or xarray.DataArray
        the xarray.DataArray needs to have the dimensions
        'cue_vector_dimensions' and 'outcome_vector_dimensions'
    n_jobs : int
        an integer giving the number of threads in which the job should be
        executed
    n_outcomes_per_job : int
        an integer giving the number of outcomes that are processed in one job
    remove_duplicates : {None, True, False}
        if None though a ValueError when the same cue is present multiple times
        in the same event; True make cues and outcomes unique per event; False
        keep multiple instances of the same cue or outcome (this is usually not
        preferred!)
    verbose : bool
        print some output if True.
    temporary_directory : str
        path to directory to use for storing temporary files created;
        if none is provided, the operating system's default will
        be used like '/tmp' on unix
    events_per_temporary_file: int
        Number of events in each temporary binary file. Has to be larger than 1

    Returns
    -------
    weights : xarray.DataArray
        with dimensions 'outcome_vector_dimensions' and
        'cue_vector_dimensions'. You can lookup the weights between a
        outcome_vector_dimension and a cue_vector_dimension with
        ``weights.loc[{'outcome_vector_dimensions': outcome_vector_dimension,
            'cue_vector_dimensions': cue_vector_dimension}]`` or
        ``weights.loc[outcome_vector_dimension].loc[cue_vector_dimension]``.

    """

    if not (remove_duplicates is None or isinstance(remove_duplicates, bool)):
        raise ValueError("remove_duplicates must be None, True or False")
    if not isinstance(events, str):
        raise ValueError("'events' need to be the path to a gzipped event file not {}".format(type(events)))

    weights_ini = weights
    wall_time_start = time.perf_counter()
    cpu_time_start = time.process_time()

    if type(cue_vectors) == dict:
        # TODO: convert dict to xarray here
        raise NotImplementedError('dicts are not supported yet.')

    if type(outcome_vectors) == dict:
        # TODO: convert dict to xarray here
        raise NotImplementedError('dicts are not supported yet.')

    if not cue_vectors.data.data.c_contiguous:
        raise ValueError('cue_vectors have to be c_contiguous')
    if not cue_vectors.dtype == np.float64:
        raise ValueError('cue_vectors have to be of dtype np.float64')

    if not outcome_vectors.data.data.c_contiguous:
        raise ValueError('outcome_vectors have to be c_contiguous')
    if not outcome_vectors.dtype == np.float64:
        raise ValueError('outcome_vectors have to be of dtype np.float64')

    # preprocessing
    n_events, cues_from_events, outcomes_from_events = count.cues_outcomes(events,
                                                                           n_jobs=n_jobs,
                                                                           verbose=verbose)

    cues_from_events = list(cues_from_events.keys())
    cues = list(cue_vectors.coords['cues'].data)
    outcomes_from_events = list(outcomes_from_events.keys())
    outcomes = list(outcome_vectors.coords['outcomes'].data)
    cue_map = OrderedDict(((cue, ii) for ii, cue in enumerate(cues)))
    outcome_map = OrderedDict(((outcome, ii) for ii, outcome in enumerate(outcomes)))

    # check for unseen outcomes in events
    if set(outcomes_from_events) - set(outcomes):
        raise ValueError("all outcomes in events need to be specified as rows in outcome_vectors")
    if set(cues_from_events) - set(cues):
        raise ValueError("all cues in events need to be specified as rows in cue_vectors")

    del outcomes_from_events, cues_from_events

    shape = (outcome_vectors.shape[1], cue_vectors.shape[1])

    if not outcome_vectors.dims[1] == 'outcome_vector_dimensions':
        raise ValueError("The second dimension of the 'outcome_vectors' has to be named 'outcome_vector_dimensions'.")
    if not cue_vectors.dims[1] == 'cue_vector_dimensions':
        raise ValueError("The second dimension of the 'cue_vectors' has to be named 'cue_vector_dimensions'.")

    cue_dims = cue_vectors['cue_vector_dimensions']
    outcome_dims = outcome_vectors['outcome_vector_dimensions']

    # initialize weights
    if weights is None:
        weights = np.ascontiguousarray(np.zeros(shape, dtype=np.float64, order='C'))
        weights = xr.DataArray(weights,
                               coords={'outcome_vector_dimensions': outcome_dims, 'cue_vector_dimensions': cue_dims},
                               dims=('outcome_vector_dimensions', 'cue_vector_dimensions'))
        del cue_dims, outcome_dims
    elif isinstance(weights, xr.DataArray):
        if not weights.shape == shape:
            raise ValueError("Vector dimensions need to match in continued learning!")
        if not all(outcome_dims == weights['outcome_vector_dimensions']):
            raise ValueError("Outcome vector dimensions names do not match in weights and outcome_vectors")
        if not all(cue_dims == weights['cue_vector_dimensions']):
            raise ValueError("Cue vector dimensions names do not match in weights and cue_vectors")
        # align the cue and outcome vector dimension names between the old weights
        # and the cue_vectors / outcome_vectors
        weights = weights.loc[{'outcome_vector_dimensions': outcome_dims, 'cue_vector_dimensions': cue_dims}]
        weights = weights.copy()
    else:
        raise ValueError('weights need to be None or xarray.DataArray with method=%s' % method)
    del shape

    if not weights.data.data.c_contiguous:
        raise ValueError('weights has to be c_contiguous')
    if method == 'numpy':
        event_generator = io.events_from_file(events)
        number_events = 0

        if verbose:
            print('start learning...')
        for cues, outcomes in event_generator:
            number_events += 1
            if verbose and number_events % 1000:
                print('.', end='')
                sys.stdout.flush()
            if remove_duplicates is None:
                if (len(cues) != len(set(cues)) or
                        len(outcomes) != len(set(outcomes))):
                    raise ValueError('cues or outcomes needs to be unique: cues '
                                     '"%s"; outcomes "%s"; use '
                                     'remove_duplicates=True' %
                                     (' '.join(cues), ' '.join(outcomes)))
            elif remove_duplicates:
                cues = set(cues)
                outcomes = set(outcomes)
            else:
                pass

            # TODO: implement multiple cues / outcomes in numpy
            assert len(outcomes) == 1, 'for method=numpy only one outcome is allowed per event'
            assert len(cues) == 1, 'for method_numpy only one cue is allowed per event'

            cue_vec = cue_vectors.loc[cues[0]]
            outcome_vec = outcome_vectors.loc[outcomes[0]]

            prediction_vec = weights.dot(cue_vec)  # why is the @ not working?
            error = outcome_vec - prediction_vec
            weights += eta * error * cue_vec  # broadcasted array multiplication
            # NOTE: we could calculate the same weights first on the first half
            # of the outcome vector dimensions and then on the second half and
            # row bind both in the end. Do we? Yes, and we can use this to
            # multiprocess the numpy computation.
    elif method in ('openmp', 'threading'):
        with tempfile.TemporaryDirectory(prefix="pyndl", dir=temporary_directory) as binary_path:
            number_events = preprocess.create_binary_event_files(events, binary_path, cue_map,
                                                                 outcome_map, overwrite=True,
                                                                 n_jobs=n_jobs,
                                                                 events_per_file=events_per_temporary_file,
                                                                 remove_duplicates=remove_duplicates,
                                                                 verbose=verbose)
            assert n_events == number_events, (str(n_events) + ' ' + str(number_events))
            binary_files = [os.path.join(binary_path, binary_file)
                            for binary_file in os.listdir(binary_path)
                            if os.path.isfile(os.path.join(binary_path, binary_file))]
            # sort binary files as they were created
            binary_files.sort(key=lambda filename: int(os.path.basename(filename)[9:-4]))
            if verbose:
                print('start learning...')
            # learning
            if method == 'openmp':
                if not sys.platform.startswith('linux'):
                    raise NotImplementedError("OpenMP is linux only at the moment."
                                              "Use method='threading' instead.")
                ndl_openmp.learn_inplace_real_to_real(binary_files,
                                                      eta,
                                                      cue_vectors.data,
                                                      outcome_vectors.data,
                                                      weights.data,
                                                      n_outcomes_per_job,
                                                      n_jobs)
            else:
                # TODO: implement threading
                raise ValueError('TODO: for now: method needs to be "numpy" or "openmp"')

            weights = weights.reset_coords(drop=True)
    else:
        raise ValueError('method needs to be "numpy" or "openmp"')

    cpu_time_stop = time.process_time()
    wall_time_stop = time.perf_counter()
    cpu_time = cpu_time_stop - cpu_time_start
    wall_time = wall_time_stop - wall_time_start

    if weights_ini is not None:
        attrs_to_be_updated = weights_ini.attrs
    else:
        attrs_to_be_updated = None

    attrs = _attributes(events, number_events, eta, cpu_time, wall_time,
                        __name__ + "." + ndl.__name__, method=method, attrs=attrs_to_be_updated)

    # post-processing
    weights.attrs = attrs
    return weights


def _attributes(event_path, number_events, eta, cpu_time,
                wall_time, function, method=None, attrs=None):

    width = max([len(ss) for ss in (event_path,
                                    str(number_events),
                                    str(eta),
                                    function,
                                    str(method),
                                    socket.gethostname(),
                                    getpass.getuser())])
    width = max(19, width)

    def _format(value):
        return '{0: <{width}}'.format(value, width=width)

    new_attrs = {'date': _format(time.strftime("%Y-%m-%d %H:%M:%S")),
                 'event_path': _format(event_path),
                 'number_events': _format(number_events),
                 'lambda': _format(str(eta)),
                 'function': _format(function),
                 'method': _format(str(method)),
                 'cpu_time': _format(str(cpu_time)),
                 'wall_time': _format(str(wall_time)),
                 'hostname': _format(socket.gethostname()),
                 'username': _format(getpass.getuser()),
                 'pyndl': _format(pyndl_version),
                 'numpy': _format(np.__version__),
                 'pandas': _format(pd.__version__),
                 'xarray': _format(xr.__version__),
                 'cython': _format(cython.__version__)}

    if attrs is not None:
        for key in set(attrs.keys()) | set(new_attrs.keys()):
            if key in attrs:
                old_val = attrs[key]
            else:
                old_val = ''
            if key in new_attrs:
                new_val = new_attrs[key]
            else:
                new_val = ''
            new_attrs[key] = old_val + ' | ' + new_val
    return new_attrs
